// Unreal Fighting Framework by Kevin Poretti

#pragma once

// FF includes
#include "FFInputBufferComponent.h"
#include "IFFSystemInterface.h"
#include "Utils/TCircleBuffer.h"
#include "State/IFFStateOwnerInterface.h"

// UE includes
#include "CoreMinimal.h"
#include "GameFramework/PlayerController.h"
#include "InputMappingContext.h"

#include "FFPlayerController.generated.h"

/**
 *	A class that collects player inputs, stores them in an input buffer, and sends a rolling window of
 *	unacknowledged inputs to a remote client or server for processing.
 */
UCLASS()
class UNREALFIGHTINGFRAMEWORK_API AFFPlayerController : public APlayerController, public IFFStateOwnerInterface
{
	GENERATED_BODY()

public:
	AFFPlayerController();

	/**
	 *	Sends all unacknowledged inputs to the remote
	 */
	virtual void SendInputsToRemote() const;

    /**
     * @brief Gets the current input after cleaning/modifying the raw input state
     * @return the current input
     */
    virtual FFFInputState GetModifiedInput() const { return ModifiedInput; }

    /**
     * @brief Function called before inputs are passed to the game logic to update the game state.
     * This is a chance for the application to mutate inputs before the game uses them to update game state.
     * Examples would include treating opposite directional inputs being held cancelling each other out or setting
     * a "neutral input" flag when no directional inputs are being held.
     * For stick/axis values this can be clamping those values or normalizing the Move and Look direction vectors.
     */
    virtual void ModifyRawInput();

    /**
     * @brief Checks if the current input state has movement input being applied
     * @return True if move axes is non-zero, false otherwise
     */
	UFUNCTION(BlueprintPure)
    bool HasMoveInput();

	/**
	 * @brief Zeroes out the raw movement axes after input has been processed 
	 */ 
	void ConsumeMoveInput();

	/**
	 * @brief Adds the current state of the player input to the input buffer for processing and
	 * interpretation after input has been cleaned
	 */
    void AddCurrentInputToBuffer();

    // IFFStateOwnerInterface
	UFUNCTION()
	virtual bool CheckInputSequence(const FFFInputSequence& InInputSequence) override;

	UFUNCTION()
    virtual bool CheckInputSequences(const TArray<FFFInputSequence>& InputSequences) override;

	virtual FVector GetInputAsWorldDirection() const override;

	virtual void DisableMostRecentInput() override;
	// End of IFFStateOwnerInterface

	// APlayerController interface
	virtual void SetupInputComponent() override;

	virtual void FlushPressedKeys() override;
	// End of APlayerController interface

protected:
    /** MappingContext */
	UPROPERTY(EditAnywhere, BlueprintReadOnly, Category = "UFF|Input", meta = (AllowPrivateAccess = "true"))
	TSoftObjectPtr<UInputMappingContext> DefaultInputMapping;

	/** Input Buffer component */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category = "UFF|Input", meta = (AllowPrivateAccess = "true"))
	UFFInputBufferComponent* InputBuffer;

	/** Current state of the player's inputs */
	FFFInputState RawInput;

	/** Current state of the player's inputs after performing cleaning and modifications */
	FFFInputState ModifiedInput;

	/**
	 *	Rolling window of the player's past inputs that have yet to be
	 *	acknowledged and simulated by the remote machine
	 *
	 *	This ring buffer should be initialized to be the size of the past X frames
	 *	you want the remote machine to re-simulate, where X is the oldest input you want to
	 *	allow to be re-simulated.
	 */
	TCircleBuffer<FFFInputState, 20> UnacknowledgedInputs;
};
